{- |

mtlstats
Copyright (C) Rhéal Lamothe
<rheal.lamothe@gmail.com>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or (at
your option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

-}

module Mtlstats.Util
  ( nth
  , modifyNth
  , dropNth
  , updateMap
  , slice
  , capitalizeName
  , linesToWidget
  , linesToWidgetC
  , userText
  ) where

import Brick.Types (Widget)
import Brick.Widgets.Center (hCenter)
import Brick.Widgets.Core (str, vBox)
import Brick.Widgets.Edit (Editor, editContentsL, getEditContents)
import Data.Char (isSpace, toUpper)
import qualified Data.Map as M
import Data.Text.Zipper (insertChar)
import Lens.Micro ((^.), (&), (%~), to)

-- | Attempt to select the element from a list at a given index
nth
  :: Int
  -- ^ The index
  -> [a]
  -- ^ The list
  -> Maybe a
nth _ [] = Nothing
nth n (x:xs)
  | n == 0    = Just x
  | n < 0     = Nothing
  | otherwise = nth (pred n) xs

-- | Attempt to modify the index at a given index in a list
modifyNth
  :: Int
  -- ^ The index
  -> (a -> a)
  -- ^ The modification function
  -> [a]
  -- ^ The list
  -> [a]
modifyNth n f = zipWith
  (\i x -> if i == n then f x else x)
  [0..]

-- | Attempt to drop the nth element from a list
dropNth
  :: Int
  -- ^ The index of the element to drop
  -> [a]
  -- ^ The list to be modified
  -> [a]
  -- ^ The modified list
dropNth n = foldr
  (\(i, x) acc -> if i == n then acc else x : acc)
  [] . zip [0..]

-- | Modify a value indexed by a given key in a map using a default
-- initial value if not present
updateMap
  :: Ord k
  => k
  -- ^ The key
  -> a
  -- ^ The default initial value
  -> (a -> a)
  -- ^ The modification function
  -> M.Map k a
  -- ^ The map
  -> M.Map k a
updateMap k def f m = let
  x = M.findWithDefault def k m
  in M.insert k (f x) m

-- | Selects a section of a list
slice
  :: Int
  -- ^ The index to start at
  -> Int
  -- ^ The number of elements to take
  -> [a]
  -- ^ The list to take a subset of
  -> [a]
slice offset len = take len . drop offset

-- | Name capitalization function for a player
capitalizeName
  :: Char
  -- ^ The character being input
  -> Editor String ()
  -- ^ The current string
  -> Editor String ()
  -- ^ The resulting string
capitalizeName ch e = e & editContentsL %~ insertChar ch'
  where
    s   = e^.to userText
    ch' = if lockFlag s
      then toUpper ch
      else ch
    lockFlag "" = True
    lockFlag (c:cs)
      | c == ','  = lockFlag' cs
      | otherwise = lockFlag cs
    lockFlag' "" = True
    lockFlag' (c:cs)
      | isSpace c = lockFlag' cs
      | otherwise = False

-- | Converts a list of lines to a widget
linesToWidget :: [String] -> Widget ()
linesToWidget = vBox . map (str . keepBlank)

-- | Converts a list of lines to a widget with each line horizontally
-- centered
linesToWidgetC :: [String] -> Widget ()
linesToWidgetC = vBox . map (hCenter . str . keepBlank)

-- | Fetches the text from an editor widget
userText :: Editor String () -> String
userText w = case getEditContents w of
  (x:_) -> x
  []    -> ""

keepBlank :: String -> String
keepBlank "" = " "
keepBlank s = s
